import sys
from pathlib import Path
from typing import Any, Callable, Iterator, List
from unittest import mock

import click
import pytest
from yarl import URL

from neuromation.api import (
    FileStatusType,
    StorageProgressComplete,
    StorageProgressDelete,
    StorageProgressEnterDir,
    StorageProgressFail,
    StorageProgressLeaveDir,
    StorageProgressStart,
    StorageProgressStep,
)
from neuromation.cli.formatters.storage import (
    BaseStorageProgress,
    DeleteProgress,
    StreamProgress,
    TTYProgress,
    create_storage_progress,
    format_url,
)
from neuromation.cli.root import Root


def unstyle(report: BaseStorageProgress) -> List[str]:
    assert isinstance(report, TTYProgress)
    return [click.unstyle(line) for (key, is_dir, line) in report.lines]


def test_format_url_storage() -> None:
    u = URL("storage://asvetlov/folder")
    assert format_url(u) == "storage://asvetlov/folder"


def test_format_url_file() -> None:
    u = URL("file:///asvetlov/folder")
    if sys.platform == "win32":
        expected = "\\asvetlov\\folder"
    else:
        expected = "/asvetlov/folder"
    assert format_url(u) == expected


_MakeRoot = Callable[[bool, bool, bool], Root]


@pytest.fixture
def make_root() -> Iterator[_MakeRoot]:
    root = None

    def make(color: bool, tty: bool, verbose: bool) -> Root:
        nonlocal root
        root = Root(
            color,
            tty,
            (80, 25),
            True,
            60,
            Path("~/.neuro"),
            verbosity=int(verbose),
            trace=False,
            trace_hide_token=True,
            command_path="",
            command_params=[],
            skip_gmp_stats=True,
            show_traceback=False,
        )
        return root

    yield make
    if root is not None:
        root.close()


def test_progress_factory_none(make_root: _MakeRoot) -> None:
    progress = create_storage_progress(make_root(False, False, False), False)
    assert isinstance(progress, StreamProgress)


def test_progress_factory_verbose(make_root: _MakeRoot) -> None:
    progress = create_storage_progress(make_root(False, False, False), False)
    assert isinstance(progress, StreamProgress)


def test_progress_factory_percent(make_root: _MakeRoot) -> None:
    progress = create_storage_progress(make_root(False, False, False), True)
    assert isinstance(progress, TTYProgress)


def test_quiet_stream_progress(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, False, False), False)
    src = URL("file:///abc")
    dst = URL("storage:xyz")

    report.begin(src, dst)
    captured = capsys.readouterr()
    assert captured.out == f""

    report.enter(StorageProgressEnterDir(src, dst))
    captured = capsys.readouterr()
    assert captured.out == f""

    report.start(StorageProgressStart(src, dst, 600))
    captured = capsys.readouterr()
    assert captured.out == f""

    report.step(StorageProgressStep(src, dst, 300, 600))
    captured = capsys.readouterr()
    assert captured.out == ""

    report.step(StorageProgressStep(src, dst, 400, 600))
    captured = capsys.readouterr()
    assert captured.out == ""

    report.complete(StorageProgressComplete(src, dst, 600))
    captured = capsys.readouterr()
    assert captured.out == f""

    report.leave(StorageProgressLeaveDir(src, dst))
    captured = capsys.readouterr()
    assert captured.out == f""


def test_stream_progress(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, False, True), False)
    src = URL("file:///abc")
    src_str = "/abc" if not sys.platform == "win32" else "\\abc"
    dst = URL("storage:xyz")
    dst_str = "storage:xyz"

    report.begin(src, dst)
    captured = capsys.readouterr()
    assert captured.out == f"Copy '{src_str}' -> '{dst_str}'\n"

    report.enter(StorageProgressEnterDir(src, dst))
    captured = capsys.readouterr()
    assert captured.out == f"'{src_str}' -> '{dst_str}'\n"

    report.start(StorageProgressStart(src, dst, 600))
    captured = capsys.readouterr()
    assert captured.out == f""

    report.step(StorageProgressStep(src, dst, 300, 600))
    captured = capsys.readouterr()
    assert captured.out == ""

    report.step(StorageProgressStep(src, dst, 400, 600))
    captured = capsys.readouterr()
    assert captured.out == ""

    report.complete(StorageProgressComplete(src, dst, 600))
    captured = capsys.readouterr()
    assert captured.out == f"'{src_str}' -> '{dst_str}'\n"

    report.leave(StorageProgressLeaveDir(src, dst))
    captured = capsys.readouterr()
    assert captured.out == f""


def test_stream_fail1(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), False)
    src = URL("file:///abc")
    src_str = "/abc" if not sys.platform == "win32" else "\\abc"
    dst = URL("storage:xyz")

    report.fail(StorageProgressFail(src, dst, "error"))
    captured = capsys.readouterr()
    assert captured.err == f"Failure: '{src_str}' -> 'storage:xyz' [error]\n"


def test_stream_fail2(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), False)
    src = URL("file:///abc")
    src_str = "/abc" if not sys.platform == "win32" else "\\abc"
    dst = URL("storage:xyz")

    report.fail(StorageProgressFail(src, dst, "error"))
    captured = capsys.readouterr()
    assert captured.err == f"Failure: '{src_str}' -> 'storage:xyz' [error]\n"


def test_tty_progress(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    src = URL("file:///abc")
    dst = URL("storage:xyz")
    src_f = URL("file:///abc/file.txt")
    dst_f = URL("storage:xyz/file.txt")

    report.begin(src, dst)
    captured = capsys.readouterr()
    assert captured.out == f"Copy 'file:///abc' => 'storage:xyz'\n"

    report.enter(StorageProgressEnterDir(src, dst))
    assert unstyle(report) == ["'file:///abc' ..."]

    report.start(StorageProgressStart(src_f, dst_f, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' [0.00%] 0B of 600B"]

    report.step(StorageProgressStep(src_f, dst_f, 300, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' [50.00%] 300B of 600B"]

    report.step(StorageProgressStep(src_f, dst_f, 400, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' [66.67%] 400B of 600B"]

    report.complete(StorageProgressComplete(src_f, dst_f, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' 600B"]

    report.leave(StorageProgressLeaveDir(src, dst))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc' DONE",
    ]


def test_tty_verbose(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, True), True)
    src = URL("file:///abc")
    dst = URL("storage:xyz")

    report.begin(src, dst)
    captured = capsys.readouterr()
    assert captured.out == f"Copy\n'file:///abc'\n=>\n'storage:xyz'\n"


def test_tty_nested(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    src = URL("file:///abc")
    dst = URL("storage:xyz")
    src_f = URL("file:///abc/file.txt")
    dst_f = URL("storage:xyz/file.txt")
    src2 = URL("file:///abc/cde")
    dst2 = URL("storage:xyz/cde")
    src2_f = URL("file:///abc/cde/file.txt")
    dst2_f = URL("storage:xyz/cde/file.txt")

    report.enter(StorageProgressEnterDir(src, dst))
    assert unstyle(report) == ["'file:///abc' ..."]

    report.start(StorageProgressStart(src_f, dst_f, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' [0.00%] 0B of 600B"]

    report.step(StorageProgressStep(src_f, dst_f, 300, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' [50.00%] 300B of 600B"]

    report.step(StorageProgressStep(src_f, dst_f, 400, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' [66.67%] 400B of 600B"]

    report.complete(StorageProgressComplete(src_f, dst_f, 600))
    assert unstyle(report) == ["'file:///abc' ...", "'file.txt' 600B"]

    report.enter(StorageProgressEnterDir(src2, dst2))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc/cde' ...",
    ]

    report.start(StorageProgressStart(src2_f, dst2_f, 800))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc/cde' ...",
        "'file.txt' [0.00%] 0B of 800B",
    ]

    report.step(StorageProgressStep(src2_f, dst2_f, 300, 800))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc/cde' ...",
        "'file.txt' [37.50%] 300B of 800B",
    ]

    report.complete(StorageProgressComplete(src2_f, dst_f, 800))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc/cde' ...",
        "'file.txt' 800B",
    ]

    report.leave(StorageProgressLeaveDir(src2, dst2))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc/cde' ...",
        "'file.txt' 800B",
        "'file:///abc/cde' DONE",
    ]

    report.leave(StorageProgressLeaveDir(src, dst))
    assert unstyle(report) == [
        "'file:///abc' ...",
        "'file.txt' 600B",
        "'file:///abc/cde' ...",
        "'file.txt' 800B",
        "'file:///abc/cde' DONE",
        "'file:///abc' DONE",
    ]


def test_fail_tty(capsys: Any, make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    src = URL("file:///abc")
    dst = URL("storage:xyz")

    report.fail(StorageProgressFail(src, dst, "error"))
    captured = capsys.readouterr()
    assert captured.err == f"Failure: 'file:///abc' -> 'storage:xyz' [error]\n"


def test_tty_fmt_url(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("storage://andrew/folder/file.txt")
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'storage://andrew/folder/file.txt'"
    )


def test_tty_fmt_storage_url_over_half(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("storage://andrew/folder0/folder1/file.txt")
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'storage://andrew/.../file.txt'"
    )


def test_tty_fmt_storage_url_over_full(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL(
        "storage://andrew/"
        + "/".join("folder" + str(i) for i in range(5))
        + "/file.txt"
    )
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=False))
        == "'storage://andrew/.../folder2/folder3/folder4/file.txt'"
    )


def test_tty_fmt_url_over_half_single_segment(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("file://" + "a" * 40)
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'file://aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'"
    )


def test_tty_fmt_url_over_half_single_segment2(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("file:///" + "a" * 40)
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'file:///aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'"
    )


def test_tty_fmt_url_over_half_long_segment(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("file:///andrew/" + "a" * 30)
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'file:///.../aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'"
    )


def test_tty_fmt_file_url_over_half(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("file:///andrew/folder0/folder1/file.txt")
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'file:///.../folder1/file.txt'"
    )


def test_tty_fmt_file_url_over_full(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL(
        "file:///andrew/" + "/".join("folder" + str(i) for i in range(5)) + "/file.txt"
    )
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=False))
        == "'file:///.../folder0/folder1/folder2/folder3/folder4/file.txt'"
    )


def test_tty_fmt_url_relative_over(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("storage:folder1/folder2/folder3/folder4/folder5")
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'storage:.../folder3/folder4/folder5'"
    )


def test_tty_fmt_url_relative_over_long_2_segments(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("storage:folder/" + "a" * 30)
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'storage:.../aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'"
    )


def test_tty_fmt_url_relative_over_single_segment(make_root: _MakeRoot) -> None:
    report = create_storage_progress(make_root(False, True, False), True)
    assert isinstance(report, TTYProgress)
    url = URL("storage:" + "a" * 35)
    assert (
        click.unstyle(report.fmt_url(url, FileStatusType.FILE, half=True))
        == "'storage:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'"
    )


def test_tty_append_files(make_root: _MakeRoot) -> None:
    with mock.patch.object(TTYProgress, "HEIGHT", 3):
        report = create_storage_progress(make_root(False, True, False), True)
        assert isinstance(report, TTYProgress)
        mock.patch.object(report, "HEIGHT", 3)
        assert report.lines == []
        report.append(URL("a"), "a")
        assert report.lines == [(URL("a"), False, "a")]
        report.append(URL("b"), "b")
        assert report.lines == [(URL("a"), False, "a"), (URL("b"), False, "b")]
        report.append(URL("c"), "c")
        assert report.lines == [
            (URL("a"), False, "a"),
            (URL("b"), False, "b"),
            (URL("c"), False, "c"),
        ]
        report.append(URL("d"), "d")
        assert report.lines == [
            (URL("b"), False, "b"),
            (URL("c"), False, "c"),
            (URL("d"), False, "d"),
        ]


def test_tty_append_dir(make_root: _MakeRoot) -> None:
    with mock.patch.object(TTYProgress, "HEIGHT", 3):
        report = create_storage_progress(make_root(False, True, False), True)
        assert isinstance(report, TTYProgress)
        mock.patch.object(report, "HEIGHT", 3)
        assert report.lines == []
        report.append(URL("a"), "a", is_dir=True)
        assert report.lines == [(URL("a"), True, "a")]
        report.append(URL("a/b"), "b")
        assert report.lines == [(URL("a"), True, "a"), (URL("a/b"), False, "b")]
        report.append(URL("a/c"), "c")
        assert report.lines == [
            (URL("a"), True, "a"),
            (URL("a/b"), False, "b"),
            (URL("a/c"), False, "c"),
        ]
        report.append(URL("a/d"), "d")
        assert report.lines == [
            (URL("a"), True, "a"),
            (URL("a/c"), False, "c"),
            (URL("a/d"), False, "d"),
        ]


def test_tty_append_second_dir(make_root: _MakeRoot) -> None:
    with mock.patch.object(TTYProgress, "HEIGHT", 3):
        report = create_storage_progress(make_root(False, True, False), True)
        assert isinstance(report, TTYProgress)
        mock.patch.object(report, "HEIGHT", 3)
        assert report.lines == []
        report.append(URL("a"), "a", is_dir=True)
        assert report.lines == [(URL("a"), True, "a")]
        report.append(URL("b"), "b")
        assert report.lines == [(URL("a"), True, "a"), (URL("b"), False, "b")]
        report.append(URL("c"), "c", is_dir=True)
        assert report.lines == [
            (URL("a"), True, "a"),
            (URL("b"), False, "b"),
            (URL("c"), True, "c"),
        ]
        report.append(URL("d"), "d")
        assert report.lines == [
            (URL("a"), True, "a"),
            (URL("c"), True, "c"),
            (URL("d"), False, "d"),
        ]


def test_delete_progress_no_color(capsys: Any, make_root: _MakeRoot) -> None:
    report = DeleteProgress(make_root(False, False, False))
    url_file = URL("storage:/abc/foo")
    url_dir = URL("storage:/abc")

    report.delete(StorageProgressDelete(url_file, is_dir=False))
    captured = capsys.readouterr()
    assert captured.out == f"Removed: '{url_file}'\n"

    report.delete(StorageProgressDelete(url_dir, is_dir=True))
    captured = capsys.readouterr()
    assert captured.out == f"Removed: '{url_dir}'\n"



import asyncio
from typing import Callable
from unittest import mock

from yarl import URL

from neuromation.api import Client, Cluster, Preset
from neuromation.cli.config import prompt_cluster


async def test_prompt_cluster(make_client: Callable[..., Client]) -> None:
    clusters = {
        "first": Cluster(
            registry_url=URL("https://registry-dev.neu.ro"),
            storage_url=URL("https://storage-dev.neu.ro"),
            users_url=URL("https://users-dev.neu.ro"),
            monitoring_url=URL("https://monitoring-dev.neu.ro"),
            secrets_url=URL("https://secrets-dev.neu.ro"),
            presets={"cpu-small": Preset(cpu=1, memory_mb=1024)},
            name="first",
        ),
        "second": Cluster(
            registry_url=URL("https://registry2-dev.neu.ro"),
            storage_url=URL("https://storage2-dev.neu.ro"),
            users_url=URL("https://users2-dev.neu.ro"),
            monitoring_url=URL("https://monitoring2-dev.neu.ro"),
            secrets_url=URL("https://secrets2-dev.neu.ro"),
            presets={"cpu-small": Preset(cpu=2, memory_mb=1024)},
            name="second",
        ),
    }

    client = make_client("https://neu.ro", clusters=clusters)

    session = mock.Mock()
    loop = asyncio.get_event_loop()
    fut = loop.create_future()
    fut.set_result("second")
    session.prompt_async.return_value = fut
    ret = await prompt_cluster(client, session=session)
    assert ret == "second"


async def test_prompt_cluster_default(make_client: Callable[..., Client]) -> None:
    clusters = {
        "first": Cluster(
            registry_url=URL("https://registry-dev.neu.ro"),
            storage_url=URL("https://storage-dev.neu.ro"),
            users_url=URL("https://users-dev.neu.ro"),
            monitoring_url=URL("https://monitoring-dev.neu.ro"),
            secrets_url=URL("https://secrets-dev.neu.ro"),
            presets={"cpu-small": Preset(cpu=1, memory_mb=1024)},
            name="first",
        ),
        "second": Cluster(
            registry_url=URL("https://registry2-dev.neu.ro"),
            storage_url=URL("https://storage2-dev.neu.ro"),
            users_url=URL("https://users2-dev.neu.ro"),
            monitoring_url=URL("https://monitoring2-dev.neu.ro"),
            secrets_url=URL("https://secrets2-dev.neu.ro"),
            presets={"cpu-small": Preset(cpu=2, memory_mb=1024)},
            name="second",
        ),
    }

    client = make_client("https://neu.ro", clusters=clusters)

    session = mock.Mock()
    loop = asyncio.get_event_loop()
    fut = loop.create_future()
    fut.set_result("")
    session.prompt_async.return_value = fut

    ret = await prompt_cluster(client, session=session)
    assert ret == "first"



import re
import shlex
import subprocess
import sys
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import click
from click.utils import make_default_short_help

from neuromation.api import ConfigError

from .root import Root
from .utils import NeuroClickMixin, Option


class InternalAlias(NeuroClickMixin, click.Command):
    ignore_unknown_options = True
    allow_interspersed_args = False
    allow_extra_args = True

    def __init__(self, name: str, alias: Dict[str, str]) -> None:
        super().__init__(name)
        assert "cmd" in alias
        self.alias = alias

    def invoke(self, ctx: click.Context) -> None:
        parent = ctx.parent
        assert parent is not None
        sub_cmd, *sub_args = shlex.split(self.alias["cmd"])
        parent_cmd = parent.command
        assert isinstance(parent_cmd, click.MultiCommand)
        cmd = parent_cmd.get_command(parent, sub_cmd)
        if cmd is None:
            ctx.fail(f'Alias {self.name} refers to unknown command "{sub_cmd}"')
        with ctx:  # type: ignore
            ctx.invoked_subcommand = self.name
            sub_ctx = cmd.make_context(self.name, sub_args + ctx.args, parent=ctx)
            with sub_ctx:  # type: ignore
                sub_ctx.command.invoke(sub_ctx)

    def get_short_help_str(self, limit: int = 45) -> str:
        txt = self.alias.get("help") or "neuro " + self.alias["cmd"]
        return make_default_short_help(txt)

    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        self.format_usage(ctx, formatter)
        formatter.write_paragraph()
        assert ctx.parent is not None
        alias_cmd = self.alias["cmd"]
        formatter.write(
            "Alias for "
            + click.style(f'"{ctx.parent.info_name} {alias_cmd}"', bold=True)
        )
        formatter.write_paragraph()

        help = self.alias.get("help")
        if help is not None:
            formatter.write_paragraph()
            formatter.write_text(help)

        self.format_options(ctx, formatter)


class ExternalAlias(NeuroClickMixin, click.Command):
    ignore_unknown_options = True
    allow_interspersed_args = False
    allow_extra_args = True

    def __init__(self, name: str, alias: Dict[str, Any]) -> None:
        assert "exec" in alias
        options = _parse_options(alias.get("options", ""))
        args = _parse_args(alias.get("args", ""))
        simplified = _validate_exec(
            alias["exec"],
            {param.name for param in options},
            {param.name for param in args},
        )
        super().__init__(name, params=options + args)
        self.alias = alias
        self.simplified = simplified

    def invoke(self, ctx: click.Context) -> None:
        if self.simplified:
            args = self._build_simplified(ctx)
        else:
            args = self._build_pattern(ctx)
        ret = subprocess.run(args)
        if ret.returncode:
            sys.exit(ret.returncode)

    def _build_simplified(self, ctx: click.Context) -> List[str]:
        cmd = self.alias["exec"]
        ret = shlex.split(cmd)
        for param in self.params:
            val = ctx.params[param.name]
            ret.extend(_process_param(param, val))
        return ret

    def _build_pattern(self, ctx: click.Context) -> List[str]:
        cmd = self.alias["exec"]
        replaces = {}
        matches = re.findall(r"{\w+}", cmd)
        for match in matches:
            name = match[1:-1]  # drop curly brackets
            name = name.lower().replace("-", "_")

            for param in self.params:
                if param.name == name:
                    break
            else:  # pragma: no cover
                # Unreachable code, _validate_exec()
                # makes sure that all param names are handled
                raise ConfigError(f'Unknown parameter {name} in "{cmd}"')
            val = ctx.params[name]
            replaces[match] = " ".join(
                shlex.quote(arg) for arg in _process_param(param, val)
            )

        for name, val in replaces.items():
            cmd = cmd.replace(name, val)

        return shlex.split(cmd)

    def get_short_help_str(self, limit: int = 45) -> str:
        txt = self.alias.get("help") or self.alias["exec"]
        return make_default_short_help(txt)

    def format_help_text(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        formatter.write_paragraph()
        alias_cmd = self.alias["exec"]
        formatter.write_text("Alias for " + click.style(f'"{alias_cmd}"', bold=True))

        help = self.alias.get("help")
        if help is not None:
            formatter.write("\n")
            formatter.write_text(help)


async def find_alias(root: Root, cmd_name: str) -> Optional[click.Command]:
    config = await root.get_user_config()
    alias = config.get("alias", {}).get(cmd_name)
    if alias is None:
        # Command not found
        return None
    if "cmd" in alias:
        return InternalAlias(cmd_name, alias)
    elif "exec" in alias:
        return ExternalAlias(cmd_name, alias)
    else:  # pragma: no cover
        # This branch is unreachable,
        # Config file validator should prevent unknown alias type
        raise click.UsageError(f"Invalid alias description type for {cmd_name}")


async def list_aliases(root: Root) -> List[click.Command]:
    config = await root.get_user_config()
    ret: List[click.Command] = []
    for cmd_name, alias in config.get("alias", {}).items():
        if "cmd" in alias:
            ret.append(InternalAlias(cmd_name, alias))
        elif "exec" in alias:
            ret.append(ExternalAlias(cmd_name, alias))
        else:  # pragma: no cover
            # This branch is unreachable,
            pass
    return ret


def _parse_options(descr: List[str]) -> List[click.Parameter]:
    ret = []
    for od in descr:
        opts = []
        is_flag = True
        metavar = None
        options, _, description = od.strip().partition("  ")
        options = options.replace(",", " ").replace("=", " ")
        for s in options.split():
            if s.startswith("--"):
                if not s[2:].replace("-", "_").isidentifier():
                    raise ConfigError(f"Cannot parse option {od}")
                opts.append(s)
            elif s.startswith("-"):
                if not s[1:].isidentifier():
                    raise ConfigError(f"Cannot parse option {od}")
                opts.append(s)
            else:
                is_flag = False
                metavar = s
                metavar = metavar.upper()
                if not metavar.isidentifier():
                    raise ConfigError(f"Cannot parse option {od}")
        description = description.strip()
        ret.append(
            Option(
                opts,
                is_flag=is_flag,
                multiple=True,
                metavar=metavar,
                help=description,
            )
        )
    return ret  # type: ignore


def _parse_args(source: str) -> List[click.Parameter]:
    ret = []
    src1 = re.sub(r"([\[\]]|\.\.\.)", r" \1 ", source)
    src2 = [s for s in re.split(r"\s+|(\S*<.*?>)", src1) if s]
    required = True
    multiple = False
    brackets = False
    arg = None
    for item in src2:
        if item == "]":
            if not brackets:
                raise ConfigError(f'Missing open bracket in "{source}"')
            if arg is None:
                raise ConfigError(f'Missing argument inside brackets in "{source}"')
            brackets = False
            required = False
        elif item == "...":
            if arg is None:
                raise ConfigError(
                    f'Ellipsis (...) should follow an argument in "{source}"'
                )
            if brackets:
                raise ConfigError(f'Ellipsis (...) inside of brackets in "{source}"')
            if multiple:
                raise ConfigError(f'Successive ellipsis (...) in "{source}"')
            multiple = True
        else:
            if arg is not None:
                if brackets:
                    raise ConfigError(f'Missing close bracket in "{source}"')
                ret.append(
                    click.Argument(
                        [arg],
                        required=required,
                        nargs=-1 if multiple else 1,
                        type=click.UNPROCESSED,
                    )
                )
                required = True
                multiple = False
                brackets = False
                arg = None
            if item == "[":
                if brackets:
                    raise ConfigError(f'Nested brackets in "{source}"')
                brackets = True
            else:
                arg = item.upper()
    if arg is not None:
        if brackets:
            raise ConfigError(f'Missing close bracket in "{source}"')
        ret.append(
            click.Argument(
                [arg],
                required=required,
                nargs=-1 if multiple else 1,
                type=click.UNPROCESSED,
            )
        )
    return ret  # type: ignore


def _validate_exec(cmd: str, options: Set[str], args: Set[str]) -> bool:
    # Return True for simplified form, False otherwise
    if args & options:
        overlapped = ",".join(args & options)
        raise ConfigError(
            "The following names are present in both "
            f"positional and optional arguments: {overlapped}"
        )
    params = args | options
    matches = re.findall(r"{\w*}", cmd)
    if not matches:
        return True
    for match in matches:
        name = match[1:-1]  # drop curly brackets
        if not name:
            raise ConfigError(f'Empty substitution is not allowed in "{cmd}"')

        if not name.isidentifier():
            raise ConfigError(f'Parameter {name} is not a valid identifier in "{cmd}"')

        if not name.islower():
            raise ConfigError(f'Parameter {name} should be lowercased in "{cmd}"')

        if name not in params:
            raise ConfigError(f'Unknown parameter {name} in "{cmd}"')

    return False


def _longest(opts: List[str]) -> str:
    # group long options first
    possible_opts = sorted(opts, key=lambda x: -len(x))
    return possible_opts[0]


def _process_param(
    param: click.Parameter, val: Union[None, str, Tuple[str]]
) -> List[str]:
    if isinstance(param, click.Argument):
        if not param.required:
            if not val:
                return []
        assert val is not None
        if param.nargs != 1:
            assert isinstance(val, tuple)
            return list(val)
        else:
            assert isinstance(val, str)
            return [val]
    elif isinstance(param, click.Option):
        if not val:
            # empty tuple
            return []
        if param.is_flag:
            # parser generates bool flags only
            assert param.is_bool_flag
            # parser doesn't allow --true / --false flags
            assert not param.secondary_opts
            vals = []
            for item in val:
                vals.append(_longest(param.opts))
            return vals
        else:
            vals = []
            for item in val:
                vals.append(_longest(param.opts))
                vals.append(item)
            return vals
    else:  # pragma: no cover
        # Unreachable branch
        raise RuntimeError(f"Unsupported parameter type {type(param)}")



from typing import Tuple

import click
import pytest

from neuromation.cli.click_types import JOB_NAME, LocalRemotePortParamType


@pytest.mark.parametrize(
    "arg,val",
    [("1:1", (1, 1)), ("1:10", (1, 10)), ("434:1", (434, 1)), ("0897:123", (897, 123))],
)
def test_local_remote_port_param_type_valid(arg: str, val: Tuple[int, int]) -> None:
    param = LocalRemotePortParamType()
    assert param.convert(arg, None, None) == val


@pytest.mark.parametrize(
    "arg",
    [
        "1:",
        "-123:10",
        "34:-65500",
        "hello:45",
        "5555:world",
        "65536:1",
        "0:0",
        "none",
        "",
    ],
)
def test_local_remote_port_param_type_invalid(arg: str) -> None:
    param = LocalRemotePortParamType()
    with pytest.raises(click.BadParameter, match=".* is not a valid port combination"):
        param.convert(arg, None, None)


class TestJobNameType:
    def test_ok(self) -> None:
        name = "a-bc-def"
        assert name == JOB_NAME.convert(name, param=None, ctx=None)

    def test_too_short(self) -> None:
        with pytest.raises(ValueError, match="Invalid job name"):
            JOB_NAME.convert("a" * 2, param=None, ctx=None)

    def test_too_long(self) -> None:
        with pytest.raises(ValueError, match="Invalid job name"):
            JOB_NAME.convert("a" * 41, param=None, ctx=None)

    @pytest.mark.parametrize(
        "name",
        [
            "abc@",  # invalid character
            "abc-DEF",  # capital letters
            "abc--def",  # two consequent hyphens
            "-abc-def",  # hyphen as the first symbol
            "abc-def-",  # hyphen as the last symbol
        ],
    )
    def test_invalid_pattern(self, name: str) -> None:
        with pytest.raises(ValueError, match="Invalid job name"):
            JOB_NAME.convert(name, param=None, ctx=None)



import re
import subprocess

import pytest

from tests.e2e import Helper


pytestmark = pytest.mark.require_admin


@pytest.mark.e2e
def test_list_clusters(helper: Helper) -> None:
    # should not fail
    helper.run_cli(["admin", "get-clusters"])


@pytest.mark.e2e
def test_list_cluster_users(helper: Helper) -> None:
    # should not fail
    helper.run_cli(["admin", "get-cluster-users"])


@pytest.mark.e2e
@pytest.mark.require_admin
def test_get_cluster_users(helper: Helper) -> None:
    captured = helper.run_cli(["admin", "get-cluster-users"])
    assert captured.err == ""

    header, *users = [re.split(" +", row.strip()) for row in captured.out.split("\n")]

    assert header == ["Name", "Role"]
    for name, role in users:
        assert role in ["admin", "manager", "user"]


@pytest.mark.e2e
def test_add_cluster_user_already_exists(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            ["admin", "add-cluster-user", helper.cluster_name, helper.username, "user"]
        )
    assert cm.value.returncode == 65
    assert (
        f"Illegal argument(s) (User '{helper.username}' already exists in cluster "
        f"'{helper.cluster_name}')" in cm.value.stderr
    )


@pytest.mark.e2e
def test_add_cluster_user_does_not_exist(helper: Helper) -> None:
    username = "some-clearly-invalid-username"
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            ["admin", "add-cluster-user", helper.cluster_name, username, "user"]
        )
    assert cm.value.returncode == 72
    assert f"User 'some-clearly-invalid-username' not found" in cm.value.stderr


@pytest.mark.e2e
def test_add_cluster_user_invalid_role(helper: Helper) -> None:
    username = "some-clearly-invalid-username"
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            ["admin", "add-cluster-user", helper.cluster_name, username, "my_role"]
        )
    assert cm.value.returncode == 2
    assert "invalid choice:" in cm.value.stderr
    assert "(choose from admin, manager, user)" in cm.value.stderr


@pytest.mark.e2e
def test_remove_cluster_user_remove_oneself(helper: Helper) -> None:
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(
            ["admin", "remove-cluster-user", helper.cluster_name, helper.username]
        )
    assert cm.value.returncode == 65
    assert (
        "Illegal argument(s) (Cluster users cannot remove themselves)"
        in cm.value.stderr
    )


@pytest.mark.e2e
def test_remove_cluster_user_does_not_exist(helper: Helper) -> None:
    username = "some-clearly-invalid-username"
    with pytest.raises(subprocess.CalledProcessError) as cm:
        helper.run_cli(["admin", "remove-cluster-user", helper.cluster_name, username])
    assert cm.value.returncode == 72
    assert f"User 'some-clearly-invalid-username' not found" in cm.value.stderr
